unit uVirusTotal;

{.$DEFINE LOG_COMMUNICATION}

interface

uses
  SysUtils, Classes;

type
  IVirusTotal = interface;
  IVirusTotalResponse = interface;
  IVirusTotalReport = interface;
  IVirusTotalReportItem = interface;
  IVirusTotalScan = interface;

  TVirusTotalProgress = reference to procedure (const Sender: IVirusTotal; const BytesTransferred, TotalBytes: int64);

  IVirusTotal = interface
    // "private"
    function GetOnDownloadProgress: TVirusTotalProgress;
    function GetOnUploadProgress: TVirusTotalProgress;
    procedure SetOnDownloadProgress(const Value: TVirusTotalProgress);
    procedure SetOnUploadProgress(const Value: TVirusTotalProgress);
    // "public"
    function GetReport(const Resource: string): IVirusTotalReport;
    function ScanFile(const FileName: string): IVirusTotalScan; overload;
    function ScanFile(Stream: TStream): IVirusTotalScan; overload;
    function Rescan(const Resource: string): IVirusTotalScan;
    property OnDownloadProgress: TVirusTotalProgress read GetOnDownloadProgress write SetOnDownloadProgress;
    property OnUploadProgress: TVirusTotalProgress read GetOnUploadProgress write SetOnUploadProgress;
  end;

  EVirusTotalError = class(Exception);
  EVirusTotalThrottlingError = class(EVirusTotalError);
  EVirusTotalPrivilegesError = class(EVirusTotalError);

  TVirusTotalResult = (vtOK, vtNotFound, vtPending, vtOther);
  TVirusTotalReportItemComparer = reference to function(const Left, Right: IVirusTotalReportItem): integer;

  IVirusTotalResponse = interface
    function GetResponseResult: TVirusTotalResult;
    function GetResponseCode: integer;
    function GetMessage: string;
    property ResponseResult: TVirusTotalResult read GetResponseResult;
    property ResponseCode: integer read GetResponseCode;
    property Message: string read GetMessage;
  end;

  IVirusTotalReport = interface(IVirusTotalResponse)
    function GetScanID: string;
    function GetScanDate: TDateTime;
    function GetResourceID: string;
    function GetMD5: string;
    function GetSHA1: string;
    function GetSHA256: string;
    function GetPermalink: string;
    function GetCount: integer;
    function GetItem(Index: integer): IVirusTotalReportItem;
    procedure Sort(Comparer: TVirusTotalReportItemComparer);
    property ScanID: string read GetScanID;
    property ScanDate: TDateTime read GetScanDate;
    property ResourceID: string read GetResourceID;
    property MD5: string read GetMD5;
    property SHA1: string read GetSHA1;
    property SHA256: string read GetSHA256;
    property Permalink: string read GetPermalink;
    property Count: integer read GetCount;
    property Items[Index: integer]: IVirusTotalReportItem read GetItem; default;
  end;

  IVirusTotalReportItem = interface
    function GetEngine: string;
    function GetEngineVersion: string;
    function GetDetected: boolean;
    function GetName: string;
    function GetUpdated: TDateTime;
    property Engine: string read GetEngine;
    property EngineVersion: string read GetEngineVersion;
    property Detected: boolean read GetDetected;
    property Name: string read GetName;
    property Updated: TDateTime read GetUpdated;
  end;

  IVirusTotalScan = interface(IVirusTotalResponse)
    function GetScanID: string;
    function GetResourceID: string;
    function GetSHA256: string;
    function GetPermalink: string;
    property ScanID: string read GetScanID;
    property ResourceID: string read GetResourceID;
    property SHA256: string read GetSHA256;
    property Permalink: string read GetPermalink;
  end;

function GetVirusTotalInstance(const ApiKey: string): IVirusTotal;

implementation

uses
  HttpSend, SynaCode, blcksock, ssl_openssl,
  MimePart,
  SuperObject,
  uGUID;

const
  ROOT_URL = 'https://www.virustotal.com/vtapi/v2';

type
  TVirusTotal = class(TInterfacedObject, IVirusTotal)
  private
    fApiKey: string;
    fHttp: THttpSend;
    fBytesToSend: int64;
    fBytesSent: int64;
    fBytesRead: int64;
    fOnDownloadProgress: TVirusTotalProgress;
    fOnUploadProgress: TVirusTotalProgress;
    function GetOnDownloadProgress: TVirusTotalProgress;
    procedure SetOnDownloadProgress(const Value: TVirusTotalProgress);
    function GetOnUploadProgress: TVirusTotalProgress;
    procedure SetOnUploadProgress(const Value: TVirusTotalProgress);
  protected
    procedure SockStatusMonitor(Sender: TObject; Reason: THookSocketReason; const Value: string);
    function SendRequest(const URL: string; Input: TStream = nil; const ContentType: string = ''): ISuperObject; overload;
    function SendRequest(const URL: string; Input: AnsiString; const ContentType: string = ''): ISuperObject; overload;
  public
    constructor Create;
    destructor Destroy; override;
    property ApiKey: string read fApiKey write fApiKey;
    function GetReport(const Resource: string): IVirusTotalReport;
    function ScanFile(const FileName: string): IVirusTotalScan; overload;
    function ScanFile(Stream: TStream): IVirusTotalScan; overload;
    function Rescan(const Resource: string): IVirusTotalScan;
    property OnDownloadProgress: TVirusTotalProgress read GetOnDownloadProgress write SetOnDownloadProgress;
    property OnUploadProgress: TVirusTotalProgress read GetOnUploadProgress write SetOnUploadProgress;
  end;

  TVirusTotalResponse = class(TInterfacedObject, IVirusTotalResponse)
  private
    fContent: ISuperObject;
    function GetResponseResult: TVirusTotalResult;
    function GetResponseCode: integer;
    function GetMessage: string;
  protected
    property Content: ISuperObject read fContent;
  public
    constructor Create(Content: ISuperObject);
    destructor Destroy; override;
    property ResponseResult: TVirusTotalResult read GetResponseResult;
    property ResponseCode: integer read GetResponseCode;
    property Message: string read GetMessage;
  end;

  TVirusTotalReport = class(TVirusTotalResponse, IVirusTotalReport)
  private
    fItems: TList;
    function GetScanID: string;
    function GetScanDate: TDateTime;
    function GetResourceID: string;
    function GetMD5: string;
    function GetSHA1: string;
    function GetSHA256: string;
    function GetPermalink: string;
    function GetCount: integer;
    function GetItem(Index: integer): IVirusTotalReportItem;
  protected
    procedure BuildItems(Scans: ISuperObject);
    procedure ClearItems;
  public
    constructor Create(Content: ISuperObject);
    destructor Destroy; override;
    procedure Sort(Comparer: TVirusTotalReportItemComparer);
    property ScanID: string read GetScanID;
    property ScanDate: TDateTime read GetScanDate;
    property ResourceID: string read GetResourceID;
    property MD5: string read GetMD5;
    property SHA1: string read GetSHA1;
    property SHA256: string read GetSHA256;
    property Permalink: string read GetPermalink;
    property Count: integer read GetCount;
    property Items[Index: integer]: IVirusTotalReportItem read GetItem;
  end;

  TVirusTotalReportItem = class(TObject, IVirusTotalReportItem)
  private
    fEngine: string;
    fEngineVersion: string;
    fDetected: boolean;
    fName: string;
    fUpdated: TDateTime;
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    function GetEngine: string;
    function GetEngineVersion: string;
    function GetDetected: boolean;
    function GetName: string;
    function GetUpdated: TDateTime;
  protected
  public
    constructor Create(const Engine: string; Content: ISuperObject);
    destructor Destroy; override;
    property Engine: string read GetEngine;
    property EngineVersion: string read GetEngineVersion;
    property Detected: boolean read GetDetected;
    property Name: string read GetName;
    property Updated: TDateTime read GetUpdated;
  end;

  TVirusTotalScan = class(TVirusTotalResponse, IVirusTotalScan)
  private
    function GetScanID: string;
    function GetResourceID: string;
    function GetSHA256: string;
    function GetPermalink: string;
  protected
  public
    property ScanID: string read GetScanID;
    property ResourceID: string read GetResourceID;
    property SHA256: string read GetSHA256;
    property Permalink: string read GetPermalink;
  end;

function GetVirusTotalInstance(const ApiKey: string): IVirusTotal;
var
  Instance: TVirusTotal;
begin
  Instance := TVirusTotal.Create;
  Result := Instance; // takes care of cleanup in case of an error
  Instance.ApiKey := ApiKey;
end;

{ TVirusTotal }

constructor TVirusTotal.Create;
begin
  inherited Create;
end;

destructor TVirusTotal.Destroy;
begin
  inherited;
end;

function TVirusTotal.GetOnDownloadProgress: TVirusTotalProgress;
begin
  Result := fOnDownloadProgress;
end;

function TVirusTotal.GetOnUploadProgress: TVirusTotalProgress;
begin
  Result := fOnUploadProgress;
end;

function TVirusTotal.GetReport(const Resource: string): IVirusTotalReport;
begin
  Result := TVirusTotalReport.Create(
    SendRequest(ROOT_URL + '/file/report'
      + '?apikey=' + {$IFDEF UNICODE} string {$ENDIF} (EncodeURLElement( {$IFDEF UNICODE} AnsiString {$ENDIF} (ApiKey)))
      + '&resource=' + {$IFDEF UNICODE} string {$ENDIF} (EncodeURLElement({$IFDEF UNICODE} AnsiString {$ENDIF} (Resource)))
    )
  );
end;

function TVirusTotal.Rescan(const Resource: string): IVirusTotalScan;
var
  Request: string;
begin
  Request := ''
      + '&apikey=' + {$IFDEF UNICODE} string {$ENDIF} (EncodeURLElement( {$IFDEF UNICODE} AnsiString {$ENDIF} (ApiKey)))
      + '&resource=' + {$IFDEF UNICODE} string {$ENDIF} (EncodeURLElement({$IFDEF UNICODE} AnsiString {$ENDIF} (Resource)))
      ;
  Result := TVirusTotalScan.Create(
    SendRequest(ROOT_URL + '/file/rescan', {$IFDEF UNICODE} AnsiString {$ENDIF} (Request))
  );
end;

function TVirusTotal.ScanFile(const FileName: string): IVirusTotalScan;
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    Result := ScanFile(Stream);
  finally
    FreeAndNil(Stream);
  end;
end;

function TVirusTotal.ScanFile(Stream: TStream): IVirusTotalScan;
var
  Request: TMemoryStream;
  Wrapper, ApiKeyField, FileField: TMimePart;
  ApiKeyRaw: RawByteString;
  Boundary: string;
begin
  ApiKeyRaw := RawByteString(ApiKey);
  Request := TMemoryStream.Create;
  try
    Boundary := '--------' + GenerateUUID;
    Wrapper := TMimePart.Create;
    try
      Wrapper.Boundary := Boundary;
      Wrapper.PrimaryCode := MP_MULTIPART;
      Wrapper.EncodingCode := ME_8BIT;
      ApiKeyField := Wrapper.AddSubPart;
      ApiKeyField.PrimaryCode := MP_TEXT;
      ApiKeyField.EncodingCode := ME_8BIT;
      ApiKeyField.Headers.Add('Content-Disposition: form-data; name="apikey"');
      ApiKeyField.DecodedLines.WriteBuffer(ApiKeyRaw[1], Length(ApiKeyRaw));
      ApiKeyField.EncodePart;
      FileField := Wrapper.AddSubPart;
      FileField.PrimaryCode := MP_BINARY;
      FileField.EncodingCode := ME_8BIT;
      FileField.Headers.Add('Content-Disposition: form-data; name="file"; filename="dummy"');
      FileField.Headers.Add('Content-Type: application/octet-stream');
      FileField.DecodedLines.CopyFrom(Stream, Stream.Size - Stream.Position);
      FileField.EncodePart;
      Wrapper.ComposeParts;
      Wrapper.Lines.SaveToStream(Request);
      Request.Position := 0;
    finally
      FreeAndNil(Wrapper);
    end;
    Result := TVirusTotalScan.Create(SendRequest(ROOT_URL + '/file/scan', Request, 'multipart/form-data; boundary=' + Boundary));
  finally
    FreeAndNil(Request);
  end;
end;

function TVirusTotal.SendRequest(const URL: string; Input: TStream; const ContentType: string): ISuperObject;
var
  Method: string;
begin
  fHttp := THttpSend.Create;
  try
    Method := 'GET';
    if Input <> nil then
    begin
      Method := 'POST';
      fHttp.Sock.ResetLastError;
      fHttp.Document.CopyFrom(Input, Input.Size - Input.Position);
      if ContentType <> '' then
        fHttp.MimeType := ContentType
      else
        fHttp.MimeType := 'application/x-www-form-urlencoded';
      {$IFDEF LOG_COMMUNICATION}
      fHttp.Document.SaveToFile('request-body.bin');
      fHttp.Document.Position := 0;
      {$ENDIF}
    end;
    fHttp.Sock.OnStatus := SockStatusMonitor;
    fBytesToSend := fHttp.Document.Size;
    fBytesSent := 0;
    fBytesRead := 0;
    if not fHttp.HTTPMethod(Method, URL) then
    begin
      if fHttp.Sock.LastError <> 0 then
        raise EVirusTotalError.Create(fHttp.Sock.LastErrorDesc)
      else
        raise EVirusTotalError.Create('Connection failed');
    end;
    {$IFDEF LOG_COMMUNICATION}
    fHttp.Document.SaveToFile('response-body.bin');
    {$ENDIF}
    if fHttp.ResultCode = 204 then
      raise EVirusTotalThrottlingError.Create('Too many requests');
    if fHttp.ResultCode = 400 then
      raise EVirusTotalError.Create('Incorrect request');
    if fHttp.ResultCode = 403 then
      raise EVirusTotalPrivilegesError.Create('Not enough privileges');
    if fHttp.ResultCode <> 200 then
      raise EVirusTotalError.CreateFmt('Unspecified error (%d)', [fHttp.ResultCode]);
    if fHttp.Document.Size = 0 then
      raise EVirusTotalError.Create('Incorrect response');
    fHttp.Document.Position := 0;
    Result := TSuperObject.ParseStream(fHttp.Document, True);
  finally
    FreeAndNil(fHttp);
  end;
end;

function TVirusTotal.SendRequest(const URL: string; Input: AnsiString; const ContentType: string): ISuperObject;
var
  Stream: TMemoryStream;
begin
  Stream := TMemoryStream.Create;
  try
    if Input <> '' then
      Stream.WriteBuffer(Input[1], Length(Input));
    Stream.Position := 0;
    Result := SendRequest(URL, Stream, ContentType);
  finally
    FreeAndNil(Stream);
  end;
end;

procedure TVirusTotal.SetOnDownloadProgress(const Value: TVirusTotalProgress);
begin
  fOnDownloadProgress := Value;
end;

procedure TVirusTotal.SetOnUploadProgress(const Value: TVirusTotalProgress);
begin
  fOnUploadProgress := Value;
end;

procedure TVirusTotal.SockStatusMonitor(Sender: TObject; Reason: THookSocketReason; const Value: string);
begin
  if Reason = HR_ReadCount then
  begin
    fBytesRead := fBytesRead + StrToInt64(Value);
    if Assigned(OnDownloadProgress) then
    begin
      if fHttp.DownloadSize <= 0 then
        OnDownloadProgress(Self, 0, 1)
      else
        OnDownloadProgress(Self, fBytesRead, fHttp.DownloadSize);
    end;
  end
  else if Reason = HR_WriteCount then
  begin
    fBytesSent := fBytesSent + StrToInt64(Value);
    if Assigned(OnUploadProgress) then
      OnUploadProgress(Self, fBytesSent, fBytesToSend);
  end;
end;

{ TVirusTotalResponse }

constructor TVirusTotalResponse.Create(Content: ISuperObject);
begin
  inherited Create;
  fContent := Content;
end;

destructor TVirusTotalResponse.Destroy;
begin
  fContent := nil;
  inherited;
end;

function TVirusTotalResponse.GetMessage: string;
begin
  Result := Content.S['verbose_msg'];
end;

function TVirusTotalResponse.GetResponseCode: integer;
begin
  Result := Content.I['response_code'];
end;

function TVirusTotalResponse.GetResponseResult: TVirusTotalResult;
begin
  case ResponseCode of
    1:   Result := vtOK;
    0:   Result := vtNotFound;
    -2:  Result := vtPending;
    else Result := vtOther
  end;
end;

{ TVirusTotalReport }

procedure TVirusTotalReport.BuildItems(Scans: ISuperObject);
var
  Iterator: TSuperObjectIter;
begin
  ClearItems;
  if ObjectFindFirst(Scans, Iterator) then
  begin
    try
      repeat
        fItems.Add(TVirusTotalReportItem.Create(Iterator.key, Iterator.val));
      until not ObjectFindNext(Iterator);
    finally
      ObjectFindClose(Iterator);
    end;
  end;
end;

procedure TVirusTotalReport.ClearItems;
var
  i: integer;
begin
  for i := 0 to Pred(fItems.Count) do
  begin
    if fItems[i] <> nil then
    begin
      TObject(fItems[i]).free;
      fItems[i] := nil;
    end;
  end;
  fItems.Clear;
end;

constructor TVirusTotalReport.Create(Content: ISuperObject);
begin
  inherited Create(Content);
  fItems := TList.Create;
  BuildItems(Content.O['scans']);
end;

destructor TVirusTotalReport.Destroy;
begin
  ClearItems;
  FreeAndNil(fItems);
  inherited;
end;

function TVirusTotalReport.GetCount: integer;
begin
  Result := fItems.Count;
end;

function TVirusTotalReport.GetItem(Index: integer): IVirusTotalReportItem;
begin
  Result := TVirusTotalReportItem(fItems[Index]);
end;

function TVirusTotalReport.GetMD5: string;
begin
  Result := Content.S['md5'];
end;

function TVirusTotalReport.GetPermalink: string;
begin
  Result := Content.S['permalink'];
end;

function TVirusTotalReport.GetResourceID: string;
begin
  Result := Content.S['resource'];
end;

function TVirusTotalReport.GetScanDate: TDateTime;
var
  sDate: string;
  Year, Month, Day, Hour, Minute, Second: integer;
  Date, Time: TDateTime;
begin
  Result := 0;
  sDate := Content.S['scan_date'];
  if sDate = '' then
    Exit;
  if Length(sDate) < 19 then
    Exit;
  if sDate[5] <> '-' then
    Exit;
  if sDate[8] <> '-' then
    Exit;
  if sDate[11] <> ' ' then
    Exit;
  if sDate[14] <> ':' then
    Exit;
  if sDate[17] <> ':' then
    Exit;
  Year := StrToIntDef(Copy(sDate, 1, 4), -1);
  Month := StrToIntDef(Copy(sDate, 6, 2), -1);
  Day := StrToIntDef(Copy(sDate, 9, 2), -1);
  Hour := StrToIntDef(Copy(sDate, 12, 2), -1);
  Minute := StrToIntDef(Copy(sDate, 15, 2), -1);
  Second := StrToIntDef(Copy(sDate, 18, 2), -1);
  if not TryEncodeDate(Year, Month, Day, Date) then
    Exit;
  if not TryEncodeTime(Hour, Minute, Second, 0, Time) then
    Time := 0;
  Result := Date + Time;
end;

function TVirusTotalReport.GetScanID: string;
begin
  Result := Content.S['scan_id'];
end;

function TVirusTotalReport.GetSHA1: string;
begin
  Result := Content.S['sha1'];
end;

function TVirusTotalReport.GetSHA256: string;
begin
  Result := Content.S['sha256'];
end;

procedure TVirusTotalReport.Sort(Comparer: TVirusTotalReportItemComparer);
begin
  if not Assigned(Comparer) then
    Exit;
  fItems.SortList(
    function (Item1, Item2: Pointer): Integer
    var
      Left, Right: IVirusTotalReportItem;
    begin
      Left := TVirusTotalReportItem(Item1);
      Right := TVirusTotalReportItem(Item2);
      Result := Comparer(Left, Right);
    end
  );
end;

{ TVirusTotalReportItem }

constructor TVirusTotalReportItem.Create(const Engine: string; Content: ISuperObject);
var
  UpdatedNum: integer;
  Updated: TDateTime;
begin
  inherited Create;
  fEngine := Engine;
  fEngineVersion := Content.S['version'];
  fDetected := Content.B['detected'];
  fName := Content.S['result'];
  fUpdated := 0;
  UpdatedNum := StrToIntDef(Content.S['update'], -1);
  if UpdatedNum > 0 then
    if TryEncodeDate(UpdatedNum div 10000, (UpdatedNum div 100) mod 100, UpdatedNum mod 100, Updated) then
      fUpdated := Updated;
end;

destructor TVirusTotalReportItem.Destroy;
begin
  inherited;
end;

function TVirusTotalReportItem.GetDetected: boolean;
begin
  Result := fDetected;
end;

function TVirusTotalReportItem.GetEngine: string;
begin
  Result := fEngine;
end;

function TVirusTotalReportItem.GetEngineVersion: string;
begin
  Result := fEngineVersion;
end;

function TVirusTotalReportItem.GetName: string;
begin
  Result := fName;
end;

function TVirusTotalReportItem.GetUpdated: TDateTime;
begin
  Result := fUpdated;
end;

function TVirusTotalReportItem.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
  if GetInterface(IID, Obj) then
    Result := 0
  else
    Result := E_NOINTERFACE;
end;

function TVirusTotalReportItem._AddRef: Integer;
begin
  Result := 0;
end;

function TVirusTotalReportItem._Release: Integer;
begin
  Result := 0;
end;

{ TVirusTotalScan }

function TVirusTotalScan.GetPermalink: string;
begin
  Result := Content.S['permalink'];
end;

function TVirusTotalScan.GetResourceID: string;
begin
  Result := Content.S['resource'];
end;

function TVirusTotalScan.GetScanID: string;
begin
  Result := Content.S['scan_id'];
end;

function TVirusTotalScan.GetSHA256: string;
begin
  Result := Content.S['sha256'];
end;

end.
